#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <elf.h>

/* Creates about the smallest possible x86_64 Linux ELF executable
   that printf "Hello World!" and exits with a 0 result. */

/* A virtual address that we pick for loading our excutable into
   memory: */
#define CODE_VIRTUAL_ADDRESS 0x40000

/* The code (generated by using `objdump` on compiled assembly shown
   in comments) */
const unsigned char code[] = {
  0xbf, 0x01, 0x00, 0x00, 0x00, // mov $0x1,%edi
  0xbe, 0x9A, 0x00, 0x04, 0x00, // mov $0x400....,%esi
  0xba, 0x0d, 0x00, 0x00, 0x00, // mov $0xd,%edx
  0xb8, 0x01, 0x00, 0x00, 0x00, // mov $0x1,%eax
  0x0f, 0x05,                   // syscall 
  0xbf, 0x01, 0x00, 0x00, 0x00, // mov    $0x0,%edi
  0xb8, 0x3c, 0x00, 0x00, 0x00, // mov    $0x3c,%eax
  0x0f, 0x05,                   // syscall
  'H', 'e', 'l', 'l', 'o', ' ',
  'W', 'o', 'r', 'l', 'd', '!',
  '\n'
};

/* We'll write a ELF header, then a program header, then the code */

int main() {
  Elf64_Ehdr ehdr;
  Elf64_Phdr phdrs[1];
  Elf64_Off code_offset = sizeof(ehdr) + sizeof(phdrs);

  memset(&ehdr, 0, sizeof(ehdr));

  ehdr.e_ident[EI_MAG0] = ELFMAG0;
  ehdr.e_ident[EI_MAG1] = ELFMAG1;
  ehdr.e_ident[EI_MAG2] = ELFMAG2;
  ehdr.e_ident[EI_MAG3] = ELFMAG3;

  ehdr.e_ident[EI_CLASS] = ELFCLASS64;
  ehdr.e_ident[EI_DATA] = ELFDATA2LSB;
  ehdr.e_ident[EI_VERSION] = EV_CURRENT;
  ehdr.e_ident[EI_OSABI] = ELFOSABI_LINUX;
  ehdr.e_ident[EI_ABIVERSION] = 1;

  ehdr.e_type = ET_EXEC;
  ehdr.e_machine = EM_X86_64;

  /* Starting point is the virtual address for the whole file
     plus the offset of our machine code within the file: */
  ehdr.e_entry = CODE_VIRTUAL_ADDRESS + code_offset;

  ehdr.e_phoff = sizeof(ehdr); /* right after the ELF header */
  ehdr.e_shoff = 0;
  
  ehdr.e_flags = 0;
  ehdr.e_ehsize = sizeof(ehdr);
  ehdr.e_phentsize = sizeof(Elf64_Phdr);
  ehdr.e_phnum = sizeof(phdrs) / sizeof(Elf64_Phdr);
  ehdr.e_shentsize = sizeof(Elf64_Shdr);
  ehdr.e_shnum = 0;
  ehdr.e_shstrndx = 0;

  /* We can use a single program header that loads our code into
     memory: */
  
  phdrs[0].p_type = PT_LOAD;
  phdrs[0].p_flags = PF_R | PF_X;
  phdrs[0].p_offset = 0;
  phdrs[0].p_vaddr = CODE_VIRTUAL_ADDRESS;
  phdrs[0].p_paddr = CODE_VIRTUAL_ADDRESS;
  phdrs[0].p_filesz = code_offset + sizeof(code);
  phdrs[0].p_memsz = code_offset + sizeof(code);
  phdrs[0].p_align = 0x1000;

  /* Write out the three parts, in order: */
  {
    int fd;
    fd = open("micro", O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IXUSR);
    write(fd, &ehdr, sizeof(ehdr));
    write(fd, phdrs, sizeof(phdrs));
    write(fd, code, sizeof(code));
    close(fd);
  }

  return 0;
}
